# -*- coding: utf-8 -*-
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Zodbpack - Pack a ZODB database"""

from __future__ import print_function
from zodbtools.util import storageFromURL, parse_tid, ashex
import ZODB.serialize
from ZODB.utils import u64
from persistent.TimeStamp import TimeStamp

import sys, time
from golang import func, defer, panic

# XXX try to avoid
import numpy as np


# XXX
def zodbpack(stor, packat, gc):
    kw = {}
    if gc is not None:
        kw['gc'] = gc

    # packat(tid) -> t(float)   (IStorage.pack requres float packtime)
    # the conversion is not exact due to potential rounding.
    # we are ok to go a bit before, but not after - so that if packat specifies
    # transaction exactly, this transaction is not pruned and stays in [packat range.
    # FIXME pack(packtime) prunes transaction specified by packtime, so maybe it should be (packat not to prune
    t = TimeStamp(packat).timeTime()
    while 1:
        packat_ = TimeStamp(*time.gmtime(t)[:5]+(t%60,)).raw()
        d = u64(packat_) - u64(packat)

        if abs(d) > 1E2:
            panic("too much divergence due to rounding:"
                  "\npackat:  %s\npackat_: %s\nδ: %d" % (ashex(packat), ashex(packat_), d))

        #print("\npackat:  %s\npackat_: %s\nδ: %d" % (ashex(packat), ashex(packat_), d))
        if d <= 0:
            break

        t = np.nextafter(t, -np.inf)


    stor.pack(t, ZODB.serialize.referencesf, **kw)


# ----------------------------------------
import getopt

summary = "pack a ZODB database"

def usage(out):
    print("""\
Usage: zodb pack [OPTIONS] <storage> <packat>
Pack a ZODB database.

<storage> is an URL (see 'zodb help zurl') of a ZODB-storage.

XXX

Options:

        --nogc      don't perform garbage-collection
    -h  --help      show this help
""", file=out)

@func
def main(argv):
    gc = True

    try:
        optv, argv = getopt.getopt(argv[1:], "h", ["nogc", "help"])
    except getopt.GetoptError as e:
        print(e, file=sys.stderr)
        usage(sys.stderr)
        sys.exit(2)

    for opt, _ in optv:
        if opt in ("-h", "--help"):
            usage(sys.stdout)
            sys.exit(0)
        if opt in ("--nogc"):
            gc = False

    if len(argv) != 2:
        usage(sys.stderr)
        sys.exit(2)

    storurl = argv[0]
    packat  = parse_tid(argv[1])

    stor = storageFromURL(storurl)
    defer(stor.close)

    zodbpack(stor, packat, gc)
